/************************************************************************/
/*                                                                      */
/*   Program Name :  usbadc2k5                                          */
/*                                                                      */
/*     Usbadc2k5 converts format from USBADC(cosmo) to K5/VSSP32        */
/*                                                                      */
/*     Enviroment variables                                             */
/*           M5DIR      ---   default directory for Mark5 data          */
/*           M5VEX      ---   default directory for Mark5 vex file      */
/*                                                                      */
/*            Ver 1.10  2015-11-05 start development                    */
/*                                                                      */
/*      Copyright (c) 2015-     T.Kondo All Right Reserved              */
/*                                                                      */
/************************************************************************/
/*

     2015-10-20  -- development started  (only 1ch mode support)
     2015-10-27  -- updated
     2015-11-05  -- add scale factor as a run parameter

           VSSP32 format #22 is newly defined to support 1000MHz sampling

   how to run
          usbadc2k5 usbadc_filename [options]
    
     where   usbadc_filename -- USBADC data file name
              options (any order)

               -ad[bit] adbit   -- # of AD conversion bit (1,2,4,8) (deflt=1)
               -fs[ampl] fmhz   -- channel sampling frequency (MHz) (deflt=same as USBADC)
                                 (frequency conversion is not implemented)
               -af fact  -- set amplitude factor (default is 1.0)
                            or threshold level in the unit of sigma for adbit=2
                            (recommended value is 0.84 for adbit=2)

               -t YYYYDDDHHMMSS|YYYY/DDD-HH:MM:SS
                  or HHMMSS|HH:MM:SS
                  or YYYYDDD|YYYY/DDD
                       -- set start date and time of data.
               -n[sample] nsample  -- set # of sampled data in a frame compulsoly
               -o k5name     -- K5 file name to be created
                                 (default : see below)
               -d k5dir      -- directory for k5 file out
                                 (default  : mark5 directory)
               -s[offset] soffset
			         +VE : start offset (sec) (default is 0)
					 -VE : absolute value means skip over bytes 
  
  Naming rule for K5 file created (USBADC : original usbadc data file name)
            
            USBADC.k5
*/

#define PROGRAM_VERSION "usbadc2k5 (Ver 1.10  2015-11-05)"

#include "usbadc2k5.h"

/* default conversion information file name set */
//#define INFONAME_DFLT "m5tok5info.txt"   // information file default name
/* environment variable for Mark5 data default directory set */
#define M5DIR "M5DIR" // the environment variable for Mark5 data default directory
/* environment variable for Mark5 vex file default directory set */
#define M5VEX "M5VEX" // the environment variable for Mark5 vex file


/* default parameters  */
#define LEN_USBADC_HEAD  192    // USBADC data heaader length (bytes)
//#define LEN_USBADC_FRAME 4288   // USBADC data frame length (bytes) for 2048 points AD (default)
//#define LEN_USBADC_FRAME 20192   // USBADC data frame length (bytes) for 10000 points AD
//#define DFLT_NUM_SAMPLES_FRAME 2048   // USBADC data # of samples data in a frame (system default value)
#define DFLT_NUM_SAMPLES_FRAME 200000   // USBADC data # of samples data in a frame for 2015/10/30 data (default value)
#define DFLT_USBADC_FSAMPL_MHZ  1000  // USBADC sampling frequency
#define DFLT_USBADC_ADBITS  16  // USBADC AD resolution
#define DFLT_USBADC_NUMCH  1  // USBADC # of channels


#define DFLT_K5_ADBITS    1     // default K5 AD# after format conversion


#define DFLT_DATA_FILE "usbadc.dat"
#define BUFMAX 25000   // maximum read buffer size

/* start date and time of data for checking purpose */
#define DFLT_YEAR  2015   // year  (for check)
#define DFLT_DDD   123    // day of year  (for check)
#define DFLT_HH    12     // hour  (for check)
#define DFLT_MM    34     // minute  (for check)
#define DFLT_SS    56     // second  (for check)


void time_option_parse(char *cbuf, int *itime);
int cmnd_get_num(int argc, char *argv[], char *command, char *param, int *num);
void param_get_runp(int argc, char *argv[],runp_t *runp);
void print_usage(void);
void monit_runp(runp_t *runp);
void vssp32head_make(vssp32_t *v32, unsigned int *head);
void vssp32head_sec_update(int seconds, unsigned int *head);
void convADres(int *buf1, int nw1, int *buf2, int ad2, int *nw2, double fact);
int convUSBADC2K5(runp_t *runp);
void spline(double *x, double *y, int npos, double *t, 
            int kpos, double *spl, double sigma);
void convSfreq(int *buf1, int nw1, int *buf2, int nw2, int *nw2t);
void get_sigma(short int *sbuf, int ndata, double *ave, double *sigma);


void time_option_parse(char *cbuf, int *itime)
{
/*
    time option parse 2010-06-09

   [in]
      cbuf --  "YYYYDDDHHMMSS|YYYY/DDD-HH:MM:SS"
                or "HHMMSS|HH:MM:SS"
                or "YYYYDDD|YYYY/DDD"
   [return]
      itime[5] --- year, day of year, hour, minute, second
                   unsatified element is not reset.
    
    
*/
    int len;
    char abuf[10];

    len=strlen(cbuf);

    if(len==17){  //  YYYY/DDD-HH:MM:SS
        strncpy(abuf,cbuf,4);    abuf[4]='\0'; itime[0]=atoi(abuf); // year
        strncpy(abuf,cbuf+5,3);  abuf[3]='\0'; itime[1]=atoi(abuf); // day of year
        strncpy(abuf,cbuf+9,2);  abuf[2]='\0'; itime[2]=atoi(abuf); // hour
        strncpy(abuf,cbuf+12,2); abuf[2]='\0'; itime[3]=atoi(abuf); // minute
        strncpy(abuf,cbuf+15,2); abuf[2]='\0'; itime[4]=atoi(abuf); // second
    } else if(len==13){ // YYYYDDDHHMMSS
        strncpy(abuf,cbuf,4);    abuf[4]='\0'; itime[0]=atoi(abuf); // year
        strncpy(abuf,cbuf+4,3);  abuf[3]='\0'; itime[1]=atoi(abuf); // day of year
        strncpy(abuf,cbuf+7,2);  abuf[2]='\0'; itime[2]=atoi(abuf); // hour
        strncpy(abuf,cbuf+9,2);  abuf[2]='\0'; itime[3]=atoi(abuf); // minute
        strncpy(abuf,cbuf+11,2); abuf[2]='\0'; itime[4]=atoi(abuf); // second
    } else if(len==8){  // HH:MM:SS or YYYY/DDD
        if(cbuf[4]=='/'){ // YYYY/DDD
           strncpy(abuf,cbuf,4);    abuf[4]='\0'; itime[0]=atoi(abuf); // year
           strncpy(abuf,cbuf+5,3);  abuf[3]='\0'; itime[1]=atoi(abuf); // day of year
        } else {   // HH:MM:SS
           strncpy(abuf,cbuf,2);    abuf[2]='\0'; itime[2]=atoi(abuf); // hour
           strncpy(abuf,cbuf+3,2);  abuf[2]='\0'; itime[3]=atoi(abuf); // minute
           strncpy(abuf,cbuf+6,2);  abuf[2]='\0'; itime[4]=atoi(abuf); // second
        }
    } else if(len==6){  // HHMMSS
        strncpy(abuf,cbuf,2);    abuf[2]='\0'; itime[2]=atoi(abuf); // hour
        strncpy(abuf,cbuf+2,2);  abuf[2]='\0'; itime[3]=atoi(abuf); // minute
        strncpy(abuf,cbuf+4,2);  abuf[2]='\0'; itime[4]=atoi(abuf); // second
    } else if(len==7){  // YYYYDDD
        strncpy(abuf,cbuf,4);    abuf[4]='\0'; itime[0]=atoi(abuf); // year
        strncpy(abuf,cbuf+4,3);  abuf[3]='\0'; itime[1]=atoi(abuf); // day of year
    }
}

int cmnd_get_num(int argc, char *argv[], char *command, char *param, int *num)
// get argument like -command parameter
// search "-command" in argument, if it is exsted, 0 is returned as
// a function value, and next argument is returned as a parameter (ascii).
// if "-command" is not exsited, -1 is returned as a function value.
// if "-command" is not accompanied by real parameter, NULL character
// is returned as param.
// 
{
    int n;
    char cbuf[80];

    param[0]='\0';

    //search command
    //printf("%s\n",command); // debug
    n=2;
    while(n <=argc) {
        strcpy(cbuf,*(argv+n-1));
        if(strncmp(cbuf,command,strlen(command))==0){  
            n++;
			*num=n;
            if(n > argc) return 0;
            strcpy(param,*(argv+n-1));
			*num++;
            return 0;
        }
        n++;
    }
    return -1;
}

void param_get_runp(int argc, char *argv[],runp_t *runp)                       
{
    int n,i;
    char cbuf[256];

	// initialize parameters

    /* common parameters */
    runp->ptype=1;
    runp->usbadc_fname[0]='\0';
    runp->k5name[0]='\0';
    runp->k5name_given[0]='\0';
    runp->soffset=0;
    runp->period=0;
    runp->odd=0;
    for(n=0; n<5; n++) runp->itimes[n]=0;
    /* for check */
    runp->itimes[0]=DFLT_YEAR;
    runp->itimes[1]=DFLT_DDD;
    runp->itimes[2]=DFLT_HH;
    runp->itimes[3]=DFLT_MM;
    runp->itimes[4]=DFLT_SS;
    /*****/

    runp->fact=1.0;

    runp->tflag=0;

    /* mode 1 parameters */
    runp->channel=0;
    runp->group=0;
    runp->reverse=0;
    runp->info_file[0]='\0';
    runp->outdir[0]='\0';
    runp->vex_file[0]='\0';
    runp->sid[0]='\0';
    runp->scan=1;
    runp->monit=0;
	runp->flag_compulsory=0;
	runp->skip_bytes=0;
	runp->k5numch=1; // 

    runp->usbadc_headlen=LEN_USBADC_HEAD;
    runp->usbadc_nsamples=DFLT_NUM_SAMPLES_FRAME;

    /* mode 2 parameters */
    runp->numch=1;
    runp->adbit =DFLT_K5_ADBITS;
    //runp->fmhz  =0.0;
    runp->fmhz  =DFLT_USBADC_FSAMPL_MHZ;
//	runp->nbitstream=0;
    for(n=0; n<32; n++) runp->ibit[n]=0;
    for(n=0; n<4; n++) runp->chtable[n]=0;
	runp->e2bit=2;

    /* 1st parameter (usbadc file name) get */
    if (argc>=2){
       strcpy(runp->usbadc_fname,*(argv+1));
	   if (strlen(*(argv+1))==1 && runp->usbadc_fname[0]=='0'){  // set default name
          strcpy(runp->usbadc_fname,DFLT_DATA_FILE);
	   }
    }

    /* command parameters get */
    if(cmnd_get(argc,argv,"-o",cbuf)==0) strcpy(runp->k5name_given,cbuf);  // K5 name
	if(cmnd_get(argc,argv,"-d",cbuf)==0){              // out dir
		strcpy(runp->outdir,cbuf);  
        /* env variable is substituted by real one */
        if((i=substitute_env_path_name(runp->outdir, cbuf))>0){
           strcpy(runp->outdir,cbuf);
        }
	}
    //if(cmnd_get_p(argc,argv,"-s",cbuf)==0) runp->soffset = atoi(cbuf);  // start offset
    if(cmnd_get(argc,argv,"-p",cbuf)==0) runp->period  = atoi(cbuf);  // period
	if(cmnd_get_p(argc,argv,"-si",cbuf)==0) strcpy(runp->sid,cbuf);  // station id
	if(cmnd_get_p(argc,argv,"-scan",cbuf)==0) {  // scan
		  runp->scan=atoi(cbuf);
          if(runp->scan < 1) runp->scan=1;
	}
	if(cmnd_get(argc,argv,"-monit",cbuf)==0) runp->monit=1;  // monitor mode on
	if(cmnd_get(argc,argv,"-1ch",cbuf)==0){    // K5 is 1ch mode
		runp->k5numch=1; 
        runp->chtable[0]=atoi(cbuf);
	}
    if(cmnd_get(argc,argv,"-fs",cbuf)==0) runp->fmhz  = atof(cbuf);  // VDIF sampling frequency
    if(cmnd_get(argc,argv,"-af",cbuf)==0) runp->fact  = atof(cbuf);  // VDIF sampling frequency
    if(cmnd_get(argc,argv,"-ns",cbuf)==0) runp->usbadc_nsamples  = atoi(cbuf);  // # of samples in a frame
    if(cmnd_get(argc,argv,"-ad",cbuf)==0) runp->adbit = atoi(cbuf);  // VDIF AD bits
    if(cmnd_get(argc,argv,"-bs",cbuf)==0) runp->nbitstream = atoi(cbuf);  // bit stream
	if(cmnd_get_num(argc,argv,"-4ch",cbuf,&n)==0){    // K5 is 4ch mode
		runp->k5numch=4; 
        runp->chtable[0]=atoi(cbuf);

        if(n > argc) goto exithere;
		runp->chtable[1]=atoi(*(argv + n-1 ));
        n++;
        if(n > argc) goto exithere;
		runp->chtable[2]=atoi(*(argv + n-1 ));
        n++;
        if(n > argc) goto exithere;
		runp->chtable[3]=atoi(*(argv + n-1 ));
	}
exithere:
	if(cmnd_get(argc,argv,"-2bit",cbuf)==0) runp->e2bit = atoi(cbuf);   // 2bit encode mode
	if(cmnd_get_p(argc,argv,"-s",cbuf)==0){
		if(atof(cbuf) >= 0.0){
		    runp->soffset = atoi(cbuf);   // second offset
		} else {
			runp->skip_bytes=-atoi(cbuf);  // skip bytes
		}
	}
    if(cmnd_get(argc,argv,"-k",cbuf)==0 ||cmnd_get(argc,argv,"-K",cbuf)==0 ){ // set compulsory mode
		runp->flag_compulsory=1;
        if(cbuf[0] != '\0'){
            if(cbuf[0] != '-') runp->flag_compulsory=atoi(cbuf);
        }
    }
    if(cmnd_get(argc,argv,"-t",cbuf)==0 ||cmnd_get(argc,argv,"-T",cbuf)==0 ){ // set time
        runp->tflag=1;
        time_option_parse(cbuf, runp->itimes);
    }

}

void print_usage(void)
{
    printf("%s\n",PROGRAM_VERSION);
    printf("\n");
    printf("    usbadc2k5 usbadc_file [options]\n");
    printf("\n");
    printf("where   usbadc_file  -- USBADC data file name\n");
    printf("      options (any order)\n");
    printf("        -o k5name     -- K5 file name to be created\n");
    printf("                          (default : see below)\n");
    printf("        -d k5dir      -- directory for k5 file out\n");
    printf("                          (default  : mark5 directory)\n");
    printf("        -t YYYYDDDHHMMSS|YYYY/DDD-HH:MM:SS\n");
    printf("           or HHMMSS|HH:MM:SS\n");
    printf("           or YYYYDDD|YYYY/DDD\n");
    printf("                      -- set start date and time of data.\n");
//    printf("        -bs nbitsream -- set Mark5B # of bit-stream (1,2,4,8,16,32)\n");
    printf("        -ad[bit] adbit  -- # of AD bit in K5 format (deflt=%d)\n",DFLT_K5_ADBITS);
    printf("        -ns[ample] nsample  -- # of samples in a frame of USBADC (deflt=%d)\n",DFLT_NUM_SAMPLES_FRAME);
    printf("        -fs[ampl] fmhz  -- sampling frequency (MHz) in K5 format (deflt=%d)\n",DFLT_USBADC_FSAMPL_MHZ);
    printf("        -af factor  -- set amplitude factor (deflt=1.0)\n");
    printf("                       or threshold level in the unit of sigma for adbit=2\n");
    printf("                       (recommended value is 0.84 for adbit=2)\n");
    printf("        -s[offset] soffset\n");
    printf("                        +VE : start offset (sec) (default is 0.0)\n");
    printf("                        -VE : absolute value means skip over bytes \n");
    printf("   Naming rule for K5 file created (USBADC : original USBADC file name)\n");
    printf("        USBADC.k5\n");
    printf("\n");
}

void monit_runp(runp_t *runp)
// run parameter monitor
{
       printf(" ---------- RUN CONDITION ---------\n");
       printf("   usbadc_fname  : %s\n",runp->usbadc_fname);
       printf("   k5name    : %s\n",runp->k5name_given);
       printf("   outdir    : %s\n",runp->outdir);
//       printf("   bitstream : %d\n",runp->nbitstream);
       printf("   fmhz      : %f\n",runp->fmhz);
       printf("   adbit     : %d\n",runp->adbit);
       printf("   k5numch   : %d\n",runp->k5numch);
//       printf("   vdif_channels : %d %d %d %d\n",runp->chtable[0],
//		   runp->chtable[1],runp->chtable[2],runp->chtable[3]);
//       printf("   2bit mode : %d\n",runp->e2bit);
//       printf("   infofile  : %s\n",runp->info_file);
//       printf("   channel   : %d\n",runp->channel);
//       printf("   group     : %d\n",runp->group);
//       printf("   soffset   : %d\n",runp->soffset);
//       printf("   period    : %d\n",runp->period);
//       printf("   odd       : %d\n",runp->odd);
//       if(runp->vex_file[0] != '\0')
//          printf("   vex_file  : %s\n",runp->vex_file);
//       if(runp->sid[0] != '\0' && runp->sid[0] !='-')
//          printf("   station_id: %s\n",runp->sid);
       printf(" -------------------------------------\n");
    printf("\n");
}

int main(int argc, char *argv[])
{
       int i;
       double telapsed;           // to measure processing time
       char fname[256];
       //char fpath[256];
       char program_version[60];
       clock_t cstart1sec, cfinish1sec;  // to calculate processing time
	   //time_t tstart, tfinish;    
       //mk5tok5_info_t mk5info;
       //gr_info_t gr; 
       runp_t runp;
       //mk5b_data_t mk5b;

       if(argc == 1 || (argc == 2 && strncmp(*(argv+1),"--help",6)==0 )){ // 2008-02-19
          print_usage();
          exit(0);
       }
       if(argc == 2 && strncmp(*(argv+1),"--ver",5)==0 ){ // 2008-02-19
          printf("\n");
          printf("%s\n",PROGRAM_VERSION);
          printf("\n");
          exit(0);
       } 
       if (argc>=2){
          if(strncmp(*(argv+1),"env",3)==0){
          printf("   Enviroment variables\n");
          printf("        %s ---   default directory for Mark5 data\n",M5DIR);
          printf("                      ( %s )\n",getenv(M5DIR));
          printf("        %s ---   default directory for Mark5 vex file\n",M5VEX);
          printf("                      ( %s )\n",getenv(M5VEX));
          printf("\n");
          exit(0);
          }
       }

       strcpy(program_version,PROGRAM_VERSION);

       /* run parameters get */
       param_get_runp(argc, argv, &runp); 

       /* out directory create if not set */

       if(runp.outdir[0]=='\0'){
          parse_dir_and_name(runp.usbadc_fname, runp.outdir, fname); // debug
          //printf("A %s\n",runp.mk5name);
          //printf("A %s\n",runp.outdir);
          //printf("A %s\n",fname);
       }


       /* Program Name Display */
       printf("************************************************************\n");
       printf("*    USBADC to K5/VSSP32 Data Format Converter             *\n");
       printf("*      %s  by  T.KONDO/NICT  *\n",program_version);
       printf("************************************************************\n");
       printf("\n");

       //monit_runp(&runp);

       cstart1sec = clock();
       //tstart= time(NULL);


	   if(runp.k5numch==1){
	       runp.group=0;
		   runp.m5ch=runp.chtable[0];
		   runp.channel=runp.m5ch;
	   } else {
	       runp.group=1;
	   }


   /********************************************/
   /* loop entry of Frequency group processing */ 
   /********************************************/

//loop_g:

       i=convUSBADC2K5(&runp);

       if(i < 0) goto exitend;

       printf("  K5 file (%s) has been created.\n\n",runp.k5name);

       //tfinish= time(NULL);
       //telapsed = difftime(tfinish,tstart);
       cfinish1sec = clock();
       telapsed = (double)(cfinish1sec - cstart1sec) / CLOCKS_PER_SEC;
       printf("Time elapsed (sec) for One obs process is %f\n",telapsed);

exitend: 
       return 0;
}

void vssp32head_make(vssp32_t *v32, unsigned int *head)
{
/*
     VSSP32 headerall make
*/
	int sfreq,fid;
    vssp32h_t v32h;

    //printf("VSSP32 headder size is %d bytes\n",sizeof(v32h));

    //v32->af_format=22; 

    /* set sync pattern 1 */
    v32h.syncptn1=0xFFFFFFFF;  

	/* set seconds from 0h UTC */
    v32h.seconds=v32->seconds;

	/* set ch flag */
	if(v32->numch==1){  // 1 ch mode
        v32h.chmode=0;
	} else {           // 4 ch mode
        v32h.chmode=1;
	}

    /* set sampling frequency index */
	sfreq=(int)(v32->fsample/1000);  // (kHz)
    fid=0;
    if( sfreq ==   100) fid=1;
    if( sfreq ==   200) fid=2;
    if( sfreq ==   500) fid=3;
    if( sfreq ==  1000) fid=4;
    if( sfreq ==  2000) fid=5;
    if( sfreq ==  4000) fid=6;
    if( sfreq ==  8000) fid=7;
    if( sfreq == 16000) fid=8;
    if( sfreq == 32000) fid=9;
    if( sfreq == 64000) fid=10;
    if( sfreq == 128000) fid=11;
    if( sfreq == 256000) fid=12;
    if( sfreq == 512000) fid=13;
    if( sfreq == 1024000) fid=14;
    if( sfreq == 2048000) fid=15;
    v32h.fsample=fid;

    /* set AD bit index */
    v32h.bitwidth=0;
    if( v32->adbit == 2 ) v32h.bitwidth=1;
    if( v32->adbit == 4 ) v32h.bitwidth=2;
    if( v32->adbit == 8 ) v32h.bitwidth=3;

    /* set sync pattern 2 */
    v32h.syncptn2=0x8C;

	/* set VSSP32 extention */

	/* set total day */
    v32h.totalday=v32->ddd;

    /* set year (2 digits) */
    v32h.year=v32->year % 100;

    /* set eflg to zero (always 0) */
    v32h.eflag=0;

    /* set aux field size (always 20)*/
    v32h.afsize=20;

    /* set minor version number  */
    v32h.minor=0;

    /* set major version number  */
    v32h.major=0;

	/* set aux field */

	/* set aux field format number */
    v32h.affmt=v32->af_format;

    if((v32h.fsample != 0) && (v32->adbit <=8) && (v32->numch == 1 || v32->numch == 4)){  
        v32h.affmt=2;
    }

    /* set filter frequency (always 0)*/
    v32h.lpf=0;

    /* Following is in case of AUX FORMAT #22 */

    /* set sampling frequency in AUX field */
	sfreq=(int)(v32->fsample/1.0e6+0.5);  // (MHz)
    v32h.fsample2=sfreq;

    /* set # of AD bits in AUX field */
    v32h.adbit=v32->adbit;

    /* set # of channel in AUX field */
    v32h.numch=v32->numch;

    /* set ASCII (14 characters) information in AUX field */
    strncpy(v32h.aux,"        USBADC",14);    // shoud be 14 byte

    /* copy to HEAD array for return */
    if(v32h.affmt == 22 ){
        memcpy(head,&v32h,32);
    } else {
        memcpy(head,&v32h,14);
        //memset(head+14,0x55,10);
        memset((char *)head+14,0x55,10);
        strncpy((char *)head+24,"  USBADC",8);
    }
}

void vssp32head_sec_update(int seconds, unsigned int *head)
{
/*
     VSSP32 header second updatee
*/
	/* set seconds from 0h UTC */
	head[1]=(head[1] & 0xFFFE0000) | (seconds & 0x1FFFF);
}


void convADres(int *buf1, int nw1, int *buf2, int ad2, int *nw2, double fact)
{
/*
   AD resolution converter 
           2015-11-05

   [in]
       buf1[nw1]  ---  raw USBADC data array (16bit AD resolution with a signed 16-bit integer)
       nw1        ---  buf1 size in the unit of 32-bit word
       ad2        ---  AD resolution of converted data ( 1 | 2 | 4 | 8 | 16 )
       fact       ---  factor for conversion   

   [return]
       buf2[nw2]  ---  array for converted data (offset binary, i.e., K5/VSSP format)
       nw2        ---  valid buf2 size in the unit of 32-bit word
*/
     int n,k;
     int w1,w2,w3,w4,w;
     int c1,c2,c3,c4,c;
     int nc,nw;
     int nkai,nfact;
     int s32;                        
     u_int32_t u32;      
     short int s16;   // -32,768 --  32,767  
     short int *sbuf;
     int mask,mask2;
     //int maskp[9]={0,0x1,0x3,0x7,0xF,0x1F,0x3F,0x7F,0xFF};  // no more use
     u_int16_t u16;    // 0 -- 65,535
     //uint8_t u8;      // 0 -- 255
     double dw;
     double ave,sigma;


     switch(ad2){
         case 1  : 
             mask=0x1;
             mask2=0x7FFFFFFF;
             break;
         case 2  :
             mask=0x3;
             mask2=0x3FFFFFFF;
             break;
         case 4  :
             mask=0xF;
             mask2=0xFFFFFFF;
             break;
         case 8  :
             mask=0xFF;
             mask2=0xFFFFFF;
             break;
         case 16  :
             mask=0xFFFF;
             mask2=0xFFFF;
             break;
         default  :
             mask=0xFFFF;
             mask2=0xFFFF;
     }


     sbuf=(short int *)buf1;
     nfact=16/ad2;

     if((nw1 % nfact) !=0) {   // return
         printf(" convADres/  array size (%d) is not adequate for this conversion!!\n",nw1);
         exit(0);  // stop
     }
     nkai=nw1*2;

     /* calc sigma */
     get_sigma(sbuf, nkai, &ave, &sigma);


     nw=0;
     nc=0;
     w1=0;
     for(n=0; n< nkai; n++){
         s32=(int)sbuf[n];
         dw=(double)s32;

         if(ad2 == 2){ // fact becomes threthold in unit of sigma
             dw-=ave;
             if(dw > sigma*fact){
                 w=3;
             } else if(dw >= 0.0){
                 w=2;
             } else if(dw >= -sigma*fact){
                 w=1;
             } else {
                 w=0;
             }
         } else {
             dw*=fact;
             dw=(double)s32*fact;
             if(dw > 32767.0) dw=32767.0;
             if(dw < -32768.0) dw=-32768.0;
             s32=(int)dw;

             s32+=32768;            // in order to convert from a signed integer to an offset binary
             u16=(u_int16_t)s32;
             w=((int)(u16 >> (16-ad2)) & mask);    // do division by bit shift according to AD resolution
         }
         w1=((w1 >> ad2) & mask2) | (w << (32-ad2));
         nc++;
         if(nc==nfact*2){
             buf2[nw]=w1;
             w1=0;
             nw++;
             nc=0;
         }
     }
     *nw2=nw;
     if(nc !=0) {
         printf(" convADres/  array size (%d) was not adequate for this conversion!!\n",nw1);
         printf(" convADres/  last 32-bit word incompleted (%d, %d)\n",nw,nc);
     }

}




int convUSBADC2K5(runp_t *runp)
/*
    USBADC to K5 conversion  2015-10-20 
	     (support only 1 ch mode)
*/
{
	unsigned int head[8];   // for VSSP32 header (2 X 32 bits)
	double fsample;  // sampling frequency

    int i,n,k;
    int iwrite;
	int numch;
	int adbit;       // AD bits in a Mark5B (1 or 2)
	int k5numch;     // # of ch in a K5 format data (1 or 4)
    int ierr;
	int frame_v32;   // total VSSP32 frame number counter
	int frame_usb;    // total USBADC frames read
	int seconds;
    int bstream;
    int64_t v32bytesec; // # of bytes in a 1sec frame of VSSP32 excepting headder
    int v32wordsec;     // # of 32bit-words in a 1sec frame of VSSP32 excepting headder
    int v32adbits;      // AD bits in VSSP32 format
    int v32word_c;      // word counter in a VSSP32 format excepting headder 
    int64_t writebytes;  // # of bytes written in a K5 frame
    int *wbuf;    // data write buffer
    int nword;
    double tsec,tsec1;
    short int *sbuf;
    FILE *fo;
    usbadc_file_t vd;
    vssp32_t v32;
    timeval_vdif_t time1st;


    printf("  convUSBADC2K5 : \n");
    ierr=0;

    /* initialize internal constants */

    //vd.fsample=1.0e9;    // 1GHz sampling
    vd.fsample=DFLT_USBADC_FSAMPL_MHZ*1.0e6;    // 1GHz sampling
    vd.numch=DFLT_USBADC_NUMCH;
    vd.adbit=DFLT_USBADC_ADBITS; 
    bstream=vd.numch*vd.adbit;   

    //vd.headlen=48;     // header length in 32byte word (=192bytes)
    //vd.headlen=LEN_USBADC_HEAD/4;     // header length in 32byte word (=192bytes)
    vd.headlen=runp->usbadc_headlen/4;     // header length in 32byte word (=192bytes)
    vd.nsample=runp->usbadc_nsamples;   // # of samples in a frame 
    vd.frame_bytes=vd.nsample*vd.adbit/8;   //  bytes frame excepting header
    vd.pframe=(double)(vd.nsample/vd.numch)/vd.fsample;   // true period (sec) of a frame 

    vd.unitp=(double)vd.nsample/vd.fsample;   // period of a frame

    v32adbits=runp->adbit;
    v32bytesec=(int64_t)(vd.fsample*(double)v32adbits/8.0 +0.5); // bytes/sec in VSSP32 form 
    v32wordsec=(int)(v32bytesec/4);    // words/sec 



    v32.adbit=v32adbits;
    v32.numch=vd.numch;
    if(runp->fmhz==0.0){
        v32.fsample=vd.fsample;
    } else {
        v32.fsample=runp->fmhz*1.0e6;
    }
/*
    if(v32.fsample != vd.fsample){ 
        printf("\n");
        printf(" Requested sampling frequency is %f MHz not 1000 MHz\n",v32.fsample/1.0e6);
        printf(" Sorry!! Current version does not support sampling frequency conversion!!\n");
        printf("\n");
        exit(0);
    }
*/
    v32.nsample=(int)(vd.nsample*v32.fsample/vd.fsample + 0.5);

    v32.af_format=22;       // AUX field format compulsoly set to 22


	strcpy(vd.fname,runp->usbadc_fname);
	vd.skip_bytes=runp->skip_bytes;
	vd.flag_compulsory=runp->flag_compulsory;

    /* set 1st frame time */
    vd.vhead.t.year=runp->itimes[0];
    vd.vhead.t.ddd=runp->itimes[1];
    vd.vhead.t.month=0;
    vd.vhead.t.mday=0;
    vd.vhead.t.hh=runp->itimes[2];
    vd.vhead.t.mm=runp->itimes[3];
    vd.vhead.t.ss=runp->itimes[4];


	vd.nword=vd.frame_bytes/4;   
    v32.nword=(int)(vd.nword*v32.fsample/vd.fsample + 0.5);

    //vd.ubuf  = (u_int32_t *)malloc(vd.nword*sizeof(u_int32_t)); //  data read buffer
    vd.ubuf  = (u_int32_t *)malloc(v32.nword*sizeof(u_int32_t)); //  data read buffer

    sbuf=(short int *)vd.ubuf;

    //wbuf     = (int *)malloc(vd.nword*sizeof(int)); //  data write buffer
    wbuf     = (int *)malloc(v32.nword*sizeof(int)); //  data write buffer

    vd.imax=1;

	vd.imax_c=0;

	// file open
	if( (vd.fin  = fopen( vd.fname, "rb" )) == NULL ){
         printf("File (%s) Open Error! \n",vd.fname);
         goto exit2return;
    }
	vd.flag_open=1;
    printf("  USBADC Data File : %s\n",vd.fname);
    printf("                     # of samples in a frame = %d\n",vd.nsample);
    printf("                     # of bytes   in a frame = %d\n",vd.frame_bytes+vd.headlen*4);

	fsample=vd.fsample;
	adbit=vd.adbit;
	numch=vd.numch;
	k5numch=runp->k5numch;

    v32.year=vd.vhead.t.year;
    v32.ddd=vd.vhead.t.ddd;
   	v32.seconds=vd.vhead.t.hh*3600+vd.vhead.t.mm*60+vd.vhead.t.ss;
 
 
	printf("\n");
    printf("  =========== conversion condition  =========== \n");
	printf("                  USBADC         VSSP32\n");
    printf("   # of CHs    :  %6d   ===>  %6d\n",vd.numch,v32.numch);
    printf("   AD bits     :  %6d   ===>  %6d  (factor=%f)\n",vd.adbit,v32.adbit,runp->fact);
    printf("   S Freq(MHz) :  %6.1f   ===>  %6.1f\n",vd.fsample/1.0e6,v32.fsample/1.0e6);
    printf("\n");
    printf("   Start Time  = %04d/%03d %02d:%02d:%02d\n",
             vd.vhead.t.year,vd.vhead.t.ddd,vd.vhead.t.hh,vd.vhead.t.mm,vd.vhead.t.ss);
    printf("\n");



//	if(vd.flag_vdif_anal==1){ // read pointer move to 1st valid header
//		fseek(vd.fin,vd.pos_1st_header,SEEK_SET); // move pointer by K5->bytesec
//	}

    /* foward startoffset seconds */
//	if(runp->soffset>0){
//		printf("  %d second forwading.....",runp->soffset);
//	    if(fwdNsecVDIF(runp->soffset, &vd) <0){
//           printf("Foward error %d\n",runp->soffset);
//           i=fclose( vd.fin ) ;
//           goto exit2return;
//        }
//	    printf("\n");
//	    /* get headder info again */
//        if(vdif_header_read_mode1(&vd)<0) goto exit2return;
//	    time1st=vd.vhead.t;
//	}

	// vdif read 

	//r_thread_id=vd.vhead.thread_id;  // set reading thread ID to 1st one


    /* output file create */

    if(runp->ptype==1){ // in case of Mode 1
       if(runp->k5name_given[0]=='\0'){
           k5name_make(runp->usbadc_fname, runp->outdir, runp->channel, 
              runp->group, runp->k5name);
	   } else {
           strcpy(runp->k5name,runp->k5name_given);
	   }
    }

    if( (fo  = fopen( runp->k5name, "wb" )) == NULL ){
       printf("  File  %s  Open Error! \n",runp->k5name);
       ierr= -2;
	   goto exit2return;
    }  
    //printf("  K5 Data File : %s\n",runp->k5name);


    printf("  K5 file (%s) is being created...\n",runp->k5name);

    /* header monitor */
			   // with Versatile ED
//    printf(" ==============================================================================\n");
//	printf("                                                             EXTENDED DATA\n");
//	printf(" HH:MM:SS  FRAME#  I  L  VER  CHS     #Bytes  C AD  TID SID       W1       W2\n");
//    printf(" ------------------------------------------------------------------------------\n");

	frame_v32=0;
    frame_usb=0;

    /* K5 initial header create */
    vssp32head_make(&v32,head);

    seconds=vd.vhead.t.hh*3600+vd.vhead.t.mm*60+vd.vhead.t.ss;
    //tsec=(double)seconds;
    vd.dlength=0.0;
    tsec=0.0;
    tsec1=tsec+1.0;
    v32word_c=0;
    //vd.nframsec=1;
//    while(1){


        writebytes=0;
		/* second loop */
        vssp32head_sec_update(v32.seconds,head);
		/* K5 header out */
        iwrite=fwrite(head,sizeof(int),8,fo);
        writebytes+=iwrite*4;

		while(1){
		   /* USBADC header read */
		   /* header read to skip */
           vd.wordread = fread( vd.ubuf, sizeof( unsigned int ), vd.headlen, vd.fin );

           /* USBADC data read */
		   vd.wordread = fread( vd.ubuf, sizeof( unsigned int ), vd.nword, vd.fin );
		   if (vd.wordread ==0 ) { // EOF
                   goto exithere; 
		   }
           vd.dlength+=vd.pframe;
           tsec+=(double)(vd.wordread*32/vd.adbit)/vd.fsample;
           v32word_c+=vd.wordread;
           frame_usb++;

           if(v32.fsample != vd.fsample){ // sampling frequency conversion

               convSfreq(vd.ubuf,vd.wordread, wbuf, vd.wordread*v32.fsample/vd.fsample, &nword);
               memcpy(vd.ubuf,wbuf,nword*sizeof(int));
               vd.wordread=nword;
           }

           if(v32word_c < v32wordsec){
               /* convert AD resolution */
               convADres(vd.ubuf,vd.wordread,wbuf,v32adbits,&nword,runp->fact);
               iwrite=fwrite(wbuf,sizeof(int),nword,fo);
               writebytes+=iwrite*4;

           } else { // new VSSP32 frame
               v32.seconds++;
               if(v32word_c==v32wordsec){
                   convADres(vd.ubuf,vd.wordread,wbuf,v32adbits,&nword,runp->fact);
                   iwrite=fwrite(wbuf,sizeof(int),nword,fo);
                   writebytes+=iwrite*4;
		           frame_v32++; 
               } else {
                   convADres(vd.ubuf,v32word_c-v32wordsec,wbuf,v32adbits,&nword,runp->fact);
                   iwrite=fwrite(wbuf,sizeof(int),nword,fo);
                   writebytes+=iwrite*4;
               }
               // new VSSP32 frame start
               writebytes=0;
               vssp32head_sec_update(seconds,head);
		       /* K5 header out */
               iwrite=fwrite(head,sizeof(int),8,fo);
               writebytes+=iwrite*4;
               if(v32word_c!=v32wordsec){
                   convADres(vd.ubuf,vd.wordread-(v32word_c-v32wordsec),wbuf,v32adbits,&nword,runp->fact);
                   iwrite=fwrite(wbuf,sizeof(int),nword,fo);
                   writebytes+=iwrite*4;
               }
           }
		}

//	}
exithere:
    if(vd.flag_open==1) i=fclose( vd.fin) ;
    vd.flag_open=0;
    // rest of data write
    i=fclose( fo ) ;
	printf("\n");
    //printf("  Data time length (sec) processed = %g\n",tsec);
    printf("  Data length (sec) processed  = %g\n",vd.dlength);
    printf("  Frames read from USBADC      = %d\n",frame_usb);
    printf("  Frames write to VSSP32       = %d + %lld/%lld\n",frame_v32,writebytes,v32bytesec+32);
	printf("\n");

exit2return:
       //printf("  K5 file (%s) has been created.\n\n",runp->k5name);
	free(vd.ubuf);
	free(wbuf);
    if(vd.flag_open==1) i=fclose(vd.fin);
    return 0;
}


/* copy from libapri_a.c  (for just in case) */

void spline(double *x, double *y, int npos, double *t, 
            int kpos, double *spl, double sigma) 
/*
  cubic spline interpolation 
    [in]
      x[npos]  --  X vector
      y[npos]  --  Y vector
      npos     --  size of x and y
      t[kpos]  --  X vector where interpolated value is calculated
      kpos     --  size of t
      sigma ------ tension (usually 1.0)
                   close to 0 => a cubic spline fit
                   >10 =>  a polynomial interpolation 
    [return]
      spl[kpos] -- interpolated Y values
*/
{
    int i,j,n, k, m, nm1, np1;
    double *yp;
    double delx1,dx1,dx2, delx2, delx12;
    double c1,c2,c3,s;
    double slpp1,slppn,deln,delnm1,delnn;
    double dels,exps,sinhs,sinhin,diag1,diagin,spdiag;
    double diag2;
    double *del1,*del2,*delss,*exps1,*expss,*sinhd1;
    double *exps2,*sinhd2,*sinhss;
    double sigmap;
    int *subs, *subs1;
    yp = (double *)malloc(2*npos*sizeof(double));  //temp storage
    del1 = (double *)malloc(kpos*sizeof(double));  //temp storage
    del2 = (double *)malloc(kpos*sizeof(double));  //temp storage
    delss = (double *)malloc(kpos*sizeof(double));  //temp storage
    sinhd1 = (double *)malloc(kpos*sizeof(double));  //temp storage
    sinhd2 = (double *)malloc(kpos*sizeof(double));  //temp storage
    sinhss = (double *)malloc(kpos*sizeof(double));  //temp storage
    exps1 = (double *)malloc(kpos*sizeof(double));  //temp storage
    exps2 = (double *)malloc(kpos*sizeof(double));  //temp storage
    expss = (double *)malloc(kpos*sizeof(double));  //temp storage
    subs = (int *)malloc(kpos*sizeof(int));  //temp storage
    subs1 = (int *)malloc(kpos*sizeof(int));  //temp storage
    delx1 = x[1]-x[0];        //1st incr
    dx1=(y[1]-y[0])/delx1;

    n=npos;
    m = kpos;
    nm1 = n-1;
    np1 = n+1;
    if (n == 2) {
        yp[0]=0.0;
        yp[1]=0.0;
    } else {
        delx2 = x[2]-x[1];
        delx12 = x[2]-x[0];
        c1 = -(delx12+delx1)/delx12/delx1;
        c2 = delx12/delx1/delx2;
        c3 = -delx1/delx12/delx2;

        slpp1 = c1*y[0]+c2*y[1]+c3*y[2];
        deln = x[nm1]-x[nm1-1];
        delnm1 = x[nm1-1]-x[nm1-2];
        delnn = x[nm1]-x[nm1-2];
        c1=(delnn+deln)/delnn/deln;
        c2=-delnn/deln/delnm1;
        c3=deln/delnn/delnm1;
        slppn = c3*y[nm1-2]+c2*y[nm1-1]+c1*y[nm1];
    }

    sigmap = sigma*nm1/(x[nm1]-x[0]);
    dels = sigmap*delx1;
    exps = exp(dels);
    sinhs = .5*(exps-1./exps);
    sinhin=1./(delx1*sinhs);
    diag1 = sinhin*(dels*0.5*(exps+1./exps)-sinhs);
    diagin = 1./diag1;
    spdiag = sinhin*(sinhs-dels);
    yp[npos]=diagin*spdiag;

    if ( n > 2 ){
        yp[0] = diagin*(dx1-slpp1);
        for (i=1; i<nm1; i++){
           delx2 = x[i+1]-x[i];
           dx2=(y[i+1]-y[i])/delx2;
           dels = sigmap*delx2;
           exps = exp(dels);
           sinhs = .5 *(exps-1./exps);
           sinhin=1./(delx2*sinhs);
           diag2 = sinhin*(dels*(.5*(exps+1./exps))-sinhs);
           diagin = 1./(diag1+diag2-spdiag*yp[n+i-1]);
           yp[i]=diagin*(dx2-dx1-spdiag*yp[i-1]);
           spdiag=sinhin*(sinhs-dels);
           yp[i+n]=diagin*spdiag;
           dx1=dx2;
           diag1=diag2;
        }
        diagin = 1.0/(diag1-spdiag*yp[n+ nm1-1]);
        yp[nm1] = diagin*(slppn-dx2-spdiag*yp[nm1-1]);
        for (i=npos-2; i >= 0; i--){
           yp[i] = yp[i]-yp[i+n]*yp[i+1];
        }
    }
    for (k=0; k<kpos; k++){
        subs[k]=nm1;
    }

    s = x[nm1]-x[0];
    sigmap = sigma*nm1/s;
    j=0;
    for (i=1; i<=nm1; i++){  
        while (x[i] > t[j]){  
           subs[j]=i;
           j++;
           if (j == m) goto done;
        }
    }
    done:;
    for (i=0; i<m; i++) {
        subs1[i]=subs[i]-1;
    }
    for (i=0; i<m; i++){
       del1[i] = t[i]-x[subs1[i]];
       del2[i] = x[subs[i]]-t[i];
       delss[i] = x[subs[i]]-x[subs1[i]];
       exps1[i]=exp(sigmap*del1[i]);
       sinhd1[i] = 0.5*(exps1[i]-1.0/exps1[i]);
       expss[i]=exp(sigmap*del2[i]);
       sinhd2[i]=.5*(expss[i]-1./expss[i]);
       expss[i] = exps1[i]*expss[i];
       sinhss[i]=.5*(expss[i]-1./expss[i]);
       spl[i]=(yp[subs[i]]*sinhd1[i]+yp[subs1[i]]*sinhd2[i])/sinhss[i]
        +((y[subs[i]]-yp[subs[i]])*del1[i]+(y[subs1[i]]-yp[subs1[i]])*del2[i])/delss[i];
    }
    free(yp);
    free(del1);
    free(del2);
    free(delss);
    free(sinhd1);
    free(sinhd2);
    free(sinhss);
    free(exps1);
    free(exps2);
    free(expss);
    free(subs);
    free(subs1);
    
}


void convSfreq(int *buf1, int nw1, int *buf2, int nw2, int *nw2t)
{
    // convert Sampling frequency according to the buffer length ratio
    // 16bit AD is assumed

    // input/output is signed integer!!

    //  2015-10-30
     int n,k;
     int w1,w2,w3,w4,w;
     int c1,c2,c3,c4,c;
     int nc,nw;
     int nkai,nfact;
     int s32;                        
     short int s16;   // -32,768 --  32,767  
     short int *sbuf1,*sbuf2;
     double *yd1,*td1,*yd2,*td2;
     int np1,np2;

     np1=2*nw1;  // # of samples
     np2=2*nw2;  // # of samples
     yd1 = (double *)malloc(np1*sizeof(double));
     td1 = (double *)malloc(np1*sizeof(double));
     yd2 = (double *)malloc(np2*sizeof(double));
     td2 = (double *)malloc(np2*sizeof(double));

     sbuf1=(short int *)buf1;
     sbuf2=(short int *)buf2;

     nw=0;
     nc=0;
     w1=0;
     for(n=0; n< np1; n++){
         s32=(int)sbuf1[n];
         yd1[n]=s32;
         td1[n]=(double)n/(double)(np1-1);
     }

     for(n=0; n< np2; n++){
         td2[n]=(double)n/(double)(np2-1);
     }

     spline(td1, yd1, np1, td2, np2, yd2, 1.0);

     for(n=0; n< np2; n++){
         sbuf2[n]=(short int)yd2[n];
     }

     *nw2t=nw2;
     free(td1);
     free(yd1);
     free(td2);
     free(yd2);

}


void get_sigma(short int *sbuf, int ndata, double *ave, double *sigma)
{
/*
     calculation of average and sigma of 16bit signed integerb data
         2015-11-05
*/
    int n;
    double dsum, dsum2, dw;
    double dave, dsigma;

    dsum=0.0;
    dsum2=0.0;
    for(n=0; n<ndata-1; n++){
        dw=(double)sbuf[n];
        dsum+=dw;
        dsum2+=(dw*dw);
    }
    dave=dsum/(double)ndata;
    dw=(dsum2-dave*dave)/(double)ndata;
    *sigma=sqrt(dw);
    *ave=dave;
}
